from ramka.views.base_view import BaseView
from ramka.views.errors import (
    default_error_handler,
    http_404_not_found,
    http_405_method_not_allowed,
)

__all__ = [
    "BaseView",
    "default_error_handler",
    "http_404_not_found",
    "http_405_method_not_allowed",
]



from inspect import isclass
from typing import Any, Callable, Dict, List, Optional, Union

from ramka.views import BaseView


class Route:
    """Route representation.

    A single route defines a path (e.g. `/users/`) and a view (e.g. UserView).

    The path should always be absolute which means it should start with a forward slash.
    It can consist of multiple arguments (e.g. `/users/{id}/`) and needs to be unique.

    The view can be either a class (that inherits from `ramka.views.BaseView`)
    or a function that accepts two arguments: the request and the response. For
    function-based views, additional argument with all supported methods can be
    specified. By default, each view only supports GET requests. For class-based views,
    the view can have multiple methods (e.g. get, post, put, delete, etc.) and the
    handler will be selected based on the request method.

    Each argument can have a type specified. For example, path `/users/{id:d}/` means
    that the argument `id` should be a decimal number. For full list of supported types
    see https://github.com/r1chardj0n3s/parse#format-specification.

    Fields:
        path (str): The path.
        view (Union[BaseView, Callable]): The view that will handle the path.
        methods (Optional[List[str]]): The HTTP methods supported by the view.
    """

    def __init__(
        self,
        path: str,
        view: Union[BaseView, Callable],
        methods: Optional[List[str]] = None,
    ):
        self.path = path
        self.view = view
        self.methods = methods or ["get", "head", "options"]

    def get_handler(self, method: Optional[str] = "get") -> Callable:
        """Get handler for the given method.

        If the view is a class, the handler will be selected based on the request
        method. If the view is a function, the handler will be the view itself.

        Arguments:
            method (str): Optional, the request method.

        Returns:
            Callable: The handler.

        Raises:
            (ValueError): If the method is not supported.
            (AttributeError): If the view is not a class or a function.
        """
        method = (method or "get").lower()

        if isclass(self.view):
            handler = getattr(self.view(), method, None)
            if handler is None:
                raise NotImplementedError(f"Method {method} is not allowed.")

            return handler

        if callable(self.view):
            if method not in self.methods:
                raise NotImplementedError(f"Method {method} is not allowed.")

            return self.view

        raise AttributeError("View is not a class or a function.")

    def __str__(self) -> str:
        return f"{self.path} -> {self.view}"

    def __repr__(self) -> str:
        return f"{self.path} -> {self.view}"


class ResolvedRoute(Route):
    """Resolved route representation.

    A route defines a path (e.g. `/users/`) and a view (e.g. UserView).

    A resolved route is a route with parameters resolved. It's used to match a
    request to a view and to generate a response using the parameters.

    Fields:
        path (str): The path.
        view (Union[BaseView, Callable]): The view that will handle the path.
        methods (Optional[List[str]]): The HTTP methods supported by the view.
        params (Dict[str, Any]): Resolved parameters.
    """

    def __init__(
        self,
        path: str,
        view: Union[BaseView, Callable],
        methods: Optional[List[str]],
        params: Dict[str, Any],
    ):
        super().__init__(path, view, methods)
        self.params = params

    @staticmethod
    def from_route(route: Route, params: Dict[str, Any]) -> "ResolvedRoute":
        """Create a resolved route from a route and parameters.

        Arguments:
            route (Route): The route.
            params (Dict[str, Any]): The parameters.
        """
        return ResolvedRoute(route.path, route.view, route.methods, params)

    def __str__(self) -> str:
        params_str = "&".join(f"{k}={v}" for k, v in self.params.items())
        return f"{self.path} ? {params_str} -> {self.view}"

    def __repr__(self) -> str:
        return f"{self.path} ? {self.params} -> {self.view}"


__all__ = ["Route", "ResolvedRoute"]



from webob import Response as BaseResponse


class Response(BaseResponse):
    """Application response implementation.

    There is no custom logic here at the moment, but there will be in the future.
    """



import os
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Optional

from whitenoise import WhiteNoise


class BaseStaticFilesEngine(ABC):
    """The base static files engine class.

    The static files engine is responsible for loading static files and rendering them.

    This class can be subclassed to implement a custom engine.
    """

    def __init__(
        self,
        app,
        root_dir: str = None,
        prefix: Optional[str] = "static/",
        **kwargs,
    ):
        if not root_dir or not os.path.isdir(root_dir):
            raise FileNotFoundError(f"{root_dir} is not a directory.")

        self._app = app
        self._root_dir = root_dir
        self._prefix = prefix
        self._kwargs = kwargs

    @abstractmethod
    def __call__(self, environ: Dict, start_response: Callable) -> Any:
        """The callable interface for the engine.

        It needs to be implemented by the subclass.
        """

    @abstractmethod
    def has_file(self, file_name: str) -> bool:
        """Check if the engine has a static file with the given name.

        Arguments:
            file_name (str): The name of the static file to check.

        Returns:
            bool: True if the engine has a static file with the given name,
                False otherwise.
        """


class WhiteNoiseEngine(BaseStaticFilesEngine):
    """The WhiteNoise static files engine class.

    The static files engine is responsible for loading static files and rendering them.
    This engine uses WhiteNoise library to load, and render static files.
    """

    _env: WhiteNoise = None

    def __init__(
        self,
        app,
        root_dir: str = None,
        prefix: Optional[str] = "static/",
        **kwargs,
    ):
        super().__init__(app, root_dir, prefix, **kwargs)

        self._env = WhiteNoise(
            self._app, root=self._root_dir, prefix=self._prefix, **self._kwargs
        )

    def __call__(self, environ: Dict, start_response: Callable) -> Any:
        """The callable interface for the engine."""
        return self._env(environ, start_response)

    def has_file(self, file_name: str) -> bool:
        """Check if the engine has a static file with the given name.

        Arguments:
            file_name (str): The name of the static file to check.

        Returns:
            bool: True if the engine has a static file with the given name,
                False otherwise.
        """
        return file_name in self._env.files


__all__ = ["BaseStaticFilesEngine", "WhiteNoiseEngine"]



from abc import ABC

from ramka.request import Request
from ramka.response import Response


class BaseView(ABC):  # pylint: disable=too-few-public-methods
    """Base class for all class-based views.

    A class-based view is a view that is defined by a class. It defines methods for all
    supported request methods.
    """

    def get(  # pylint: disable=no-self-use,unused-argument
        self, request: Request, response: Response, **kwargs
    ) -> None:
        """GET HTTP method handler.

        It should be overridden in subclasses to enable support for the GET HTTP method.

        Arguments:
            request (Request): The request.
            response (Response): The response.
            kwargs (dict): The route parameters.
        """
        raise NotImplementedError()

    def post(  # pylint: disable=no-self-use,unused-argument
        self, request: Request, response: Response, **kwargs
    ) -> None:
        """POST HTTP method handler.

        It should be overridden in subclasses to enable support for the POST HTTP
        method.

        Arguments:
            request (Request): The request.
            response (Response): The response.
            kwargs (dict): The route parameters.
        """
        raise NotImplementedError()

    def put(  # pylint: disable=no-self-use,unused-argument
        self, request: Request, response: Response, **kwargs
    ) -> None:
        """PUT HTTP method handler.

        It should be overridden in subclasses to enable support for the PUT HTTP method.

        Arguments:
            request (Request): The request.
            response (Response): The response.
            kwargs (dict): The route parameters.
        """
        raise NotImplementedError()

    def patch(  # pylint: disable=no-self-use,unused-argument
        self, request: Request, response: Response, **kwargs
    ) -> None:
        """PATCH HTTP method handler.

        It should be overridden in subclasses to enable support for the PATCH HTTP
        method.

        Arguments:
            request (Request): The request.
            response (Response): The response.
            kwargs (dict): The route parameters.
        """
        raise NotImplementedError()

    def delete(  # pylint: disable=no-self-use,unused-argument
        self, request: Request, response: Response, **kwargs
    ) -> None:
        """DELETE HTTP method handler.

        It should be overridden in subclasses to enable support for the DELETE HTTP
        method.

        Arguments:
            request (Request): The request.
            response (Response): The response.
            kwargs (dict): The route parameters.
        """
        raise NotImplementedError()


__all__ = ["BaseView"]



from webob import Request as BaseRequest


class Request(BaseRequest):
    """Application request implementation.

    There is no custom logic here at the moment, but there will be in the future.
    """



from requests import Session as RequestsSession
from wsgiadapter import WSGIAdapter as RequestsWSGIAdapter

from ramka.app import App
from ramka.response import Response


class TestSession(RequestsSession):
    """The test session class.

    The test session is a wrapper around the requests library. It provides helper
    methods to make testing more straightforward.
    """

    def __init__(self, base_url: str, adapter: RequestsWSGIAdapter):
        """Initialize the test session.

        Arguments:
            base_url (str): The base URL to use for the test session.
            adapter (WSGIAdapter): The adapter to use for the test session.
        """
        super().__init__()
        self._base_url = base_url
        self.mount(prefix=self._base_url, adapter=adapter)

    @property
    def base_url(self) -> str:
        """The base URL of the test session.

        Returns:
            str: The base URL of the test session.
        """
        return self._base_url.rstrip("/")

    def _request(self, method: str, url: str, **kwargs) -> Response:
        """The request method.

        It makes a request to the test session.

        Arguments:
            method (str): The request method.
            url (str): The URL to request.
            kwargs (Dict): The keyword arguments to pass to the request method.

        Returns:
            Response: The response of the request.
        """
        func = getattr(super(), method)
        if not func:
            raise AttributeError(f"Method {method} not supported.")

        url = url.lstrip("/")
        return func(f"{self.base_url}/{url}", **kwargs)

    def get(self, url, **kwargs) -> Response:
        """Perform a GET request.

        Arguments:
            url (str): The URL to request.
            kwargs (Dict): The keyword arguments to pass to the request method.

        Returns:
            Response: The response of the request.
        """
        return self._request("get", url, **kwargs)

    def post(self, url, data=None, json=None, **kwargs) -> Response:
        """Perform a POST request.

        Arguments:
            url (str): The URL to request.
            data (Optional[Dict]): The data to send with the request.
            json (Optional[Dict]): The JSON to send with the request.
            kwargs (Dict): The keyword arguments to pass to the request method.

        Returns:
            Response: The response of the request.
        """
        return self._request("post", url, data=data, json=json, **kwargs)

    def put(self, url, data=None, **kwargs) -> Response:
        """Perform a PUT request.

        Arguments:
            url (str): The URL to request.
            data (Optional[Dict]): The data to send with the request.
            kwargs (Dict): The keyword arguments to pass to the request method.

        Returns:
            Response: The response of the request.
        """
        return self._request("put", url, data=data, **kwargs)

    def patch(self, url, data=None, **kwargs) -> Response:
        """Perform a PATCH request.

        Arguments:
            url (str): The URL to request.
            data (Optional[Dict]): The data to send with the request.
            kwargs (Dict): The keyword arguments to pass to the request method.

        Returns:
            Response: The response of the request.
        """
        return self._request("patch", url, data=data, **kwargs)

    def delete(self, url, **kwargs) -> Response:
        """Perform a DELETE request.

        Arguments:
            url (str): The URL to request.
            kwargs (Dict): The keyword arguments to pass to the request method.

        Returns:
            Response: The response of the request.
        """
        return self._request("delete", url, **kwargs)


def create_test_app(app: App, base_url: str = "http://testserver") -> App:
    """Create a test app.

    It add a test session to the app existing app object to make it usable for testing.
    The test session is be added as a property called `test_session`.
    """

    def test_session(_) -> TestSession:
        """Return the test session.

        Returns:
            TestSession: The test session.
        """
        return TestSession(base_url, RequestsWSGIAdapter(app))

    setattr(type(app), "test_session", property(test_session))

    return app



from abc import ABC, abstractmethod
from typing import Callable, List, Optional, Union

from parse import parse

from ramka.routing.route import ResolvedRoute, Route
from ramka.views import BaseView


class BaseRouter(ABC):
    """Base router class.

    The responsibility of the routes is to know how to handle given request.

    Router, on the other hand, is responsible for storing all routes that have been
    defined in the application and finding correct routes for the requests.

    There are two ways to add a new route to the router: by using the `add_route` method
    or by using the `route` decorator. It is the case to allow developers adding routes
    in a preffered way.

    Fields:
        routes (List[Route]): The list of routes that have been defined in the
            application.
    """

    def __init__(self) -> None:
        self.routes: List[Route] = []

    @abstractmethod
    def add_route(
        self,
        path: str,
        view: Union[BaseView, Callable],
        methods: Optional[List[str]] = None,
    ) -> None:
        """Add a route to the router.

        It's supposed to be used as a method.

        Arguments:
            path (str): The path to add the route to.
            view (Union[BaseView, Callable]): The view to add the route to.
            methods (Optional[List[str]]): The list of methods to add the route to.
        """

    @abstractmethod
    def route(self, path: str) -> Callable:
        """Add a route to the router.

        It's supposed to be used as a decorator.

        Arguments:
            path (str): The path to add the route to.
        """

    @abstractmethod
    def resolve(self, path: str) -> ResolvedRoute:
        """Resolve the route for the given path.

        Arguments:
            path (str): The path to resolve the route for.

        Returns:
            ResolvedRoute: The resolved route.
        """

    @abstractmethod
    def has_route(self, path: str) -> bool:
        """Check if the router has a route for the given path.

        Arguments:
            path (str): The path to check.

        Returns:
            bool: True if the router has a route for the given path, False otherwise.
        """


class SimpleRouter(BaseRouter):
    """Simple router class.

    The responsibility of the routes is to know how to handle given request.

    Router, on the other hand, is responsible for storing all routes that have been
    defined in the application and finding correct routes for the requests.

    This router can resolve routes with or without trailing slashes (that behavior can
    be disabled).

    Fields:
        routes (List[Route]): The list of routes that have been defined in the
            application.
    """

    def __init__(self, force_trailing_slashes: bool = True):
        super().__init__()
        self._force_trailing_slashes = force_trailing_slashes

    def _handle_trailing_slashes(self, path: str) -> str:
        """Handle trailing slashes.

        If the router is configured to force trailing slashes, the path is updated to
        have trailing slash even if it doesn't have one.

        Otherwise, the path is returned as is.

        Arguments:
            path (str): The path to handle.

        Returns:
            str: The path with trailing slash added if necessary.
        """
        if self._force_trailing_slashes and not path.endswith("/"):
            return f"{path}/"

        return path

    def resolve(self, path: str) -> ResolvedRoute:
        """Resolve the route for the given path.

        Arguments:
            path (str): The path to resolve the route for.

        Returns:
            ResolvedRoute: The resolved route.
        """
        path = self._handle_trailing_slashes(path)
        for route in self.routes:
            parsed_path = parse(route.path, path)
            if parsed_path:
                return ResolvedRoute.from_route(route, parsed_path.named)

        return None

    def has_route(self, path: str) -> bool:
        """Check if the router has a route for the given path.

        Arguments:
            path (str): The path to check.

        Returns:
            bool: True if the router has a route for the given path, False otherwise.
        """
        return self.resolve(path) is not None

    def add_route(
        self,
        path: str,
        view: Union[BaseView, Callable],
        methods: Optional[List[str]] = None,
    ) -> None:
        """Add a route to the router.

        It's supposed to be used as a method.

        Arguments:
            path (str): The path to add the route to.
            view (Union[BaseView, Callable]): The view to add the route to.
            methods (Optional[List[str]]): The list of methods to add the route to.
        """
        if self.has_route(path):
            raise AttributeError("Route already exists.")

        self.routes.append(Route(self._handle_trailing_slashes(path), view, methods))

    def route(self, path: str, methods: Optional[List[str]] = None) -> Callable:
        """Add a route to the router.

        It's supposed to be used as a decorator.

        Arguments:
            path (str): The path to add the route to.
            methods (Optional[List[str]]): The list of methods to add the route to.

        Returns:
            Callable: The decorated function.
        """

        def wrapper(view: Union[BaseView, Callable]):
            self.add_route(path, view, methods)
            return view

        return wrapper


__all__ = ["BaseRouter", "SimpleRouter"]



from typing import TYPE_CHECKING, Type

from ramka.request import Request
from ramka.response import Response

if TYPE_CHECKING:
    from ramka.app import App


class Middleware:
    """Base middleware class.

    Middleware is a class that can be used to modify the request and response objects.
    The idea is that middleware wraps the application and it can be wrapped with another
    middleware. This way, the application can be easily extended. For example,
    a middleware can be used to log the requests, and another middleware can be used
    to add some header to the response.

    Subclass this class and override the `process_request` and `process_response`
    methods to implement custom middleware. Each middleware can override one of those
    methods or both.
    """

    def __init__(self, app: "App") -> None:
        """Initialize the middleware.

        Arguments:
            app (App): The application to wrap.
        """
        self._app = app

    def __call__(self, environ, start_response) -> Response:
        request = Request(environ)
        response = self._app.handle_request(request)

        return response(environ, start_response)

    def add(self, middleware_cls: Type["Middleware"]):
        """Add another middleware to the execution chain.

        Arguments:
            middleware_cls (Type[Middleware]): The middleware class to add.
        """
        self._app = middleware_cls(self._app)

    def handle_request(self, request: Request) -> Response:
        """Handle the request.

        This method is called by the application. It calls the `process_request`
        method before the request is handled, and then the `process_response` method
        after the request is handled.

        Arguments:
            request (Request): The request to handle.
        """
        self.process_request(request)
        response = self._app.handle_request(request)
        self.process_response(request, response)

        return response

    def process_request(self, request: Request) -> None:  # pylint: disable=no-self-use
        """Process the request.

        This method is called by the application before the request is handled. It
        can be overridden to add custom logic to the request.

        Arguments:
            request (Request): The request to process.
        """

    def process_response(  # pylint: disable=no-self-use
        self, request: Request, response: Response
    ) -> None:
        """Process the response.

        This method is called by the application after the request is handled. It
        can be overridden to add custom logic to the response. It takes the request
        as one of the arguments but modifying it won't have any effect as the request
        is already handled.

        Arguments:
            request (Request): The request object.
            response (Response): The response to process.
        """



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ramka',
 'ramka.middleware',
 'ramka.request',
 'ramka.response',
 'ramka.routing',
 'ramka.static',
 'ramka.templates',
 'ramka.test',
 'ramka.views']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'WebOb>=1.8.7,<2.0.0',
 'gunicorn>=20.1.0,<21.0.0',
 'parse>=1.19.0,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'whitenoise>=6.0.0,<7.0.0']

setup_kwargs = {
    'name': 'ramka',
    'version': '0.1.1',
    'description': 'ramka - a simple Python web framework',
    'long_description': '# ramka\n\n[![Tests](https://github.com/mateuszcisek/ramka/actions/workflows/tests.yaml/badge.svg)](https://github.com/mateuszcisek/ramka/actions/workflows/tests.yaml)\n[![Linting](https://github.com/mateuszcisek/ramka/actions/workflows/linting.yaml/badge.svg)](https://github.com/mateuszcisek/ramka/actions/workflows/linting.yaml)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://pycqa.github.io/isort/)\n[![linting: pylint](https://img.shields.io/badge/linting-pylint-yellowgreen)](https://github.com/PyCQA/pylint)\n\n\nramka is a very small web framework written in Python.\n\n## Documentation\n\nComing soon :)\n\n## Installation\n\nComing soon :)\n\n## License\n\nMIT\n',
    'author': 'Mateusz Cisek',
    'author_email': 'cismat@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mateuszcisek/ramka',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import os
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from jinja2 import Environment, FileSystemLoader, Template
from jinja2.exceptions import TemplateNotFound


def _find_template_directories(root_dir: str, template_dir_name: str) -> List[str]:
    """Find the directories that possibly contain templates.

    This function does not check if the directories actually contain templates. It only
    finds directories that with name matching the given `template_dir_name`.

    Arguments:
        root_dir (str): The root directory to start searching from.
        template_dir_name (str): The name of the directories to find.

    Returns:
        (List[str]): The list of directories that possibly contain templates.
    """
    result = []
    for rootdir, dirs, _ in os.walk(root_dir):
        for subdir in dirs:
            if subdir == template_dir_name:
                result.append(os.path.join(rootdir, subdir))

    return result


class BaseTemplateEngine(ABC):
    """The base template engine class.

    The template engine is responsible for loading templates and rendering them.

    This class can be subclassed to implement a custom template engine.
    """

    def __init__(
        self, root_dir: str = None, template_dir_name: Optional[str] = "templates"
    ):
        if not root_dir or not os.path.isdir(root_dir):
            raise FileNotFoundError(f"{root_dir} is not a directory.")

        self._root_dir = root_dir
        self._template_dir_name = template_dir_name

    @abstractmethod
    def render(self, template_name: str, context: Optional[Dict] = None) -> Any:
        """Render a template.

        Information provided in the `context` dictionary will be available in the
        template.

        Arguments:
            template_name (str): The name of the template to render.
            context (Optional[Dict]): The context to use when rendering the template.

        Returns:
            Any: The rendered template.
        """

    @abstractmethod
    def has_template(self, template_name: str) -> bool:
        """Check if the template engine has a template with the given name.

        Arguments:
            template_name (str): The name of the template to check.

        Returns:
            bool: True if the template engine has a template with the given name,
                False otherwise.
        """


class JinjaTemplateEngine(BaseTemplateEngine):
    """The Jinja template engine class.

    The template engine is responsible for loading templates and rendering them. This
    engine uses Jinja2 library to load, parse, and render templates.
    """

    def __init__(
        self, root_dir: str = None, template_dir_name: Optional[str] = "templates"
    ):
        super().__init__(root_dir, template_dir_name)

        self._env = Environment(
            loader=FileSystemLoader(
                _find_template_directories(root_dir, template_dir_name)
            )
        )

    def render(self, template_name: str, context: Optional[Dict] = None) -> Template:
        """Render a template.

        Information provided in the `context` dictionary will be available in the
        template.

        Arguments:
            template_name (str): The name of the template to render.
            context (Optional[Dict]): The context to use when rendering the template.

        Returns:
            Template: The rendered template.
        """
        if context is None:
            context = {}

        try:
            return self._env.get_template(template_name).render(**context).encode()
        except TemplateNotFound as ex:
            raise FileNotFoundError(f"{template_name} is not found.") from ex

    def has_template(self, template_name: str) -> bool:
        """Check if the template engine has a template with the given name.

        Arguments:
            template_name (str): The name of the template to check.

        Returns:
            bool: True if the template engine has a template with the given name,
                False otherwise.
        """
        return template_name in self._env.loader.list_templates()


__all__ = ["BaseTemplateEngine", "JinjaTemplateEngine"]



import json

from ramka.response import Response


def _error_page(response: Response, error_code: int, message: str) -> None:
    """Helper function to send an error page to the client.

    It updates the response object with the error details.

    Arguments:
        response (Response): The response object to send the error page to.
        error_code (int): The error code to send to the client.
        message (str): The message to send to the client.
    """
    response.status_code = error_code
    response.content_type = "application/json"
    response.text = json.dumps({"error": message})


def http_404_not_found(_, response):
    """The default 404 handler.

    It sends a 404 error page to the client.
    """
    _error_page(response, 404, "Not found.")


def http_405_method_not_allowed(_, response):
    """The default handler for methods that have not been implemented."""
    _error_page(response, 405, "Method not allowed.")


def default_error_handler(_, response, error: Exception):
    """The default handler for all unhandled exceptions.

    It sends a 500 error page to the client as all unhandles exceptions are considered
    as server errors. Instead of sending a genetic 500 error page, it sends a JSON
    response with the error message.
    """
    _error_page(response, 500, str(error))


__all__ = ["default_error_handler", "http_404_not_found", "http_405_method_not_allowed"]



from typing import Any, Callable, Dict, List, Optional, Type, Union

from ramka.middleware import Middleware
from ramka.request import Request
from ramka.response import Response
from ramka.routing import BaseRouter, SimpleRouter
from ramka.static import BaseStaticFilesEngine, WhiteNoiseEngine
from ramka.templates import BaseTemplateEngine, JinjaTemplateEngine
from ramka.views import (
    BaseView,
    default_error_handler,
    http_404_not_found,
    http_405_method_not_allowed,
)


class App:
    """The main application class.

    This is the entrypoint for the application.
    """

    def __init__(
        self,
        root_dir: str,
        *,
        router: Optional[BaseRouter] = None,
        router_kwargs: Optional[Dict] = None,
        template_engine: Optional[BaseTemplateEngine] = None,
        template_engine_kwargs: Optional[Dict] = None,
        static_files_dir: Optional[str] = None,
        static_files_engine: Optional[BaseStaticFilesEngine] = None,
        static_files_engine_kwargs: Optional[Dict] = None,
        http_404_not_found_handler: Optional[Callable] = None,
        http_405_method_not_allowed_handler: Optional[Callable] = None,
        error_handler: Optional[Callable] = None,
        middleware_classes: Optional[List[Type[Middleware]]] = None,
    ):
        """Initialize the application.

        Arguments:
            root_dir (str): The root directory of the application.
            router (Optional[BaseRouter]): The router to use.
            router_kwargs (Optional[Dict]): The kwargs to pass to the router.
            template_engine (Optional[BaseTemplateEngine]): The template engine to use.
            template_engine_kwargs (Optional[Dict]): The kwargs to pass to the template
                engine.
            static_files_dir (Optional[str]): The directory containing static files.
            static_files_engine (Optional[BaseStaticFilesEngine]): The static files
                engine to use.
            static_files_engine_kwargs (Optional[Dict]): The kwargs to pass to the
                static files engine.
            http_404_not_found_handler (Optional[Callable]): The handler to use for
                HTTP 404 error (Not found).
            http_405_method_not_allowed_handler (Optional[Callable]): The handler to use
                for HTTP 405 error (Method not allowed).
            error_handler (Optional[Callable]): The handler to use for errors.
            middleware_classes (Optional[List[Type[Middleware]]]): The list of
                middleware classes to use.

        """
        self._router = router or SimpleRouter(**(router_kwargs or {}))
        self._template_engine = self._initialize_template_engine(
            root_dir, template_engine, template_engine_kwargs
        )
        self._static_files_engine = self._initialize_static_files_engine(
            static_files_dir, static_files_engine, static_files_engine_kwargs
        )
        self._middleware = self._initialize_middleware(middleware_classes)

        self._http_404_handler = http_404_not_found_handler or http_404_not_found
        self._http_405_handler = (
            http_405_method_not_allowed_handler or http_405_method_not_allowed
        )
        self._error_handler = error_handler or default_error_handler

    def __call__(self, environ, start_response):
        if self._static_files_engine:
            return self._static_files_engine(environ, start_response)

        return self._wsgi_app(environ, start_response)

    def _initialize_template_engine(  # pylint: disable=no-self-use
        self,
        root_dir: str,
        engine: Optional[BaseTemplateEngine] = None,
        arguments: Optional[Dict] = None,
    ) -> BaseTemplateEngine:
        """Initialize the template engine.

        Arguments:
            root_dir (str): The root directory of the application.
            engine (Optional[BaseTemplateEngine]): The engine to use.
            arguments (Optional[Dict]): The kwargs to pass to the engine.

        Returns:
            BaseTemplateEngine: The initialized template engine.
        """
        return engine or JinjaTemplateEngine(**(arguments or {"root_dir": root_dir}))

    def _initialize_static_files_engine(
        self,
        static_files_dir: str,
        engine: Optional[BaseStaticFilesEngine] = None,
        arguments: Optional[Dict] = None,
    ) -> BaseStaticFilesEngine:
        """Initialize the static files engine.

        Arguments:
            static_files_dir (str): The directory containing static files.
            engine (Optional[BaseStaticFilesEngine]): The engine to use.
            arguments (Optional[Dict]): The kwargs to pass to the engine.

        Returns:
            BaseStaticFilesEngine: The initialized static files engine.
        """
        if not static_files_dir:
            return None

        return engine or WhiteNoiseEngine(
            **(arguments or {"app": self._wsgi_app, "root_dir": static_files_dir})
        )

    def _initialize_middleware(
        self, middleware_classes: List[Type[Middleware]]
    ) -> Middleware:
        """Initialize the middleware.

        Arguments:
            middleware_classes (List[Type[Middleware]]): The list of middleware classes
                to use.

        Returns:
            Middleware: The initialized middleware.
        """
        middleware = Middleware(self)
        if middleware_classes:
            for middleware_class in middleware_classes:
                middleware.add(middleware_class)

        return middleware

    def _wsgi_app(self, environ, start_response):
        """The WSGI application.

        Arguments:
            environ (Dict): The WSGI environment.
            start_response (Callable): The WSGI start response function.

        Returns:
            The response body.
        """
        request = Request(environ)
        response = self._handle_request(request)
        return response(environ, start_response)

    def _handle_request(self, request: Request) -> Response:
        """Handle a request.

        Arguments:
            request (Request): The request to handle.

        Returns:
            Response: The response.

        Raises:
            Exception: An error occurred if no handler found.
        """
        response = Response()
        parsed_route = self._router.resolve(request.path)

        try:
            if parsed_route:
                handler = parsed_route.get_handler(request.method)
                handler(request, response, **parsed_route.params)
            else:
                self._http_404_handler(request, response)

        except NotImplementedError:
            self._http_405_handler(request, response)

        # Using `Exception` class as we want to catch all exception here.
        except Exception as error:  # pylint: disable=broad-except
            if self._error_handler is None:
                raise error

            self._error_handler(request, response, error)

        return response

    def has_route(self, path: str) -> bool:
        """Check if the router has a route for the given path.

        Arguments:
            path (str): The path to check.

        Returns:
            bool: True if the router has a route for the given path, False otherwise.
        """
        return self._router.has_route(path)

    def add_route(
        self,
        path: str,
        view: Union[BaseView, Callable],
        methods: Optional[List[str]] = None,
    ) -> None:
        """Add a route to the router.

        It's supposed to be used as a method.

        Arguments:
            path (str): The path to add the route to.
            view (Union[BaseView, Callable]): The view to add the route to.
            methods (Optional[List[str]]): The list of methods to add the route to.
        """
        self._router.add_route(path, view, methods)

    def route(self, path: str, methods: Optional[List[str]] = None) -> Callable:
        """Add a route to the router.

        It's supposed to be used as a decorator.

        Arguments:
            path (str): The path to add the route to.
            methods (Optional[List[str]]): The list of methods to add the route to.
        """
        return self._router.route(path, methods)

    def template(self, template_name, context: Dict[str, Any] = None) -> Any:
        """Render a template using defined template engine.

        Arguments:
            template_name (str): The name of the template to render.
            context (Optional[Dict[str, Any]]): The context to use.

        Returns:
            Any: The rendered template.
        """
        return self._template_engine.render(template_name, context)
